#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <sys/stat.h>
#include <sys/types.h>
#include <type_traits>
#include <iostream>

#include <dune/common/exceptions.hh>
#include <dune/common/parametertree.hh>
#include <dune/common/parametertreeparser.hh>

#include <dune/grid/common/mcmgmapper.hh>
#include <dune/grid/io/file/dgfparser/dgfyasp.hh>
#include <dune/grid/io/file/dgfparser/dgfug.hh>
#include <dune/grid/yaspgrid.hh>

#include <dune/mmesh/mmesh.hh>

#include <dune/mmdg/dg.hh>
#include <dune/mmdg/clockhelper.hh>
#include <dune/mmdg/problems/dgproblem.hh>
#include <dune/mmdg/problems/dgproblem1.hh>
#include <dune/mmdg/problems/dgproblem2.hh>
#include <dune/mmdg/problems/dgproblem3.hh>

int main(int argc, char** argv)
{
  try
  {
    using Clock = std::chrono::steady_clock;

    //start run time measurement
    const Clock::time_point timeInitial = Clock::now();

    static constexpr int dim = GRIDDIM;

    //use YaspGrid if dim = 1 and MMesh otherwise
    using Grid = std::conditional<dim == 1, Dune::YaspGrid<dim>,
      Dune::MovingMesh<dim>>::type;
    using GridFactory = Dune::DGFGridFactory<Grid>;
    using GridView = typename Grid::LeafGridView;
    using Mapper = typename Dune::MultipleCodimMultipleGeomTypeMapper<GridView>;
    using Coordinate = Dune::FieldVector<double, dim>;
    using Problem = DGProblem<Coordinate>;
    using DG = DG<GridView, Mapper, Problem>;

    //get parameters from file "parameterDG.ini"
    Dune::ParameterTree pt;
    Dune::ParameterTreeParser::readINITree("parameterDG.ini", pt);

    //assume constant penalty parameter
    const double mu0 = std::stod(pt["mu0"]);

    //create a grid from .dgf file
    GridFactory gridFactory( "grids/cube" + std::to_string(dim) + "d.dgf" );

    const Grid& grid = *gridFactory.grid();
    const GridView& gridView = grid.leafGridView();

    //element mapper for the leaf grid view
    const Mapper mapper(gridView, Dune::mcmgElementLayout());

    Problem* problem; //problem to be solved

    //determine problem type
    if (pt["problem"] == "dg1")
    {
      problem = new DGProblem1<Coordinate>();
    }
    else if (pt["problem"] == "dg2")
    {
      problem = new DGProblem2<Coordinate>();
    }
    else if (pt["problem"] == "dg3")
    {
      problem = new DGProblem3<Coordinate>();

      if constexpr (dim == 2)
      {
        std::cout << "NOTE: The exact solution of your problem is given by "
          "an infinite series; increase the cut-off if necessary!\n\n";
      }
    }
    else
    {
      DUNE_THROW(Dune::Exception,
        "Invalid problem type or parameter 'problem' not specified in file "
        "parameterDG.ini (use 'dg1' to 'dg3')");
    }

    DG dg(gridView, mapper, *problem);

    dg(mu0);

    //error and total run time
    const double error = dg.computeL2error();
    const double timeTotal = ClockHelper::elapsedRuntime(timeInitial);

    //determine maximum edge length
    double largestEdgeLength = 0.;
    for (const auto& edge : edges(gridView))
    {
      largestEdgeLength = std::max(largestEdgeLength, edge.geometry().volume());
    }

    //write error, runtime and maximum edge length
    std::ofstream errorFile("convergenceData", std::ios::out | std::ios::trunc);

    if (errorFile.is_open())
    {
      errorFile << error << "\n" << timeTotal << "\n" << largestEdgeLength;
      errorFile.close();
    }
    else
    {
      std::cout << "Unable to write error data to file.\n";
    }

    //print total run time and error
    std::cout << "Finished with a total run time of " << timeTotal <<
      " Seconds.\nL2-error: " << error << ".\n\n";

    return 0;
  }
  catch (Dune::Exception& e)
  {
    std::cerr << "Dune reported error: " << e.what() << std::endl;
  }
  catch (std::exception& e)
  {
    std::cerr << "Error: " << e.what() << std::endl;
  }
  catch (...)
  {
    std::cerr << "Unknown exception thrown!" << std::endl;
  }
}
