import numpy as np
from os import mkdir, remove
from os.path import join, exists
from subprocess import call
from math import log


# changes the parameter file such that gridfile is used
def writeINI(gridfile):
    # storage for file content
    lines = None

    # new line to write
    newline = "gridfile = " + gridfile

    # indicates whether the parameter gridfile already exists
    hasGridlineParameter = False

    # read file content and replace gridfile parameter if existent
    with open("parameterMMDG.ini", "r") as file:
        lines = file.read().splitlines()

        for i in range(len(lines)):
            if (lines[i].startswith("gridfile")):
                lines[i] = newline
                hasGridlineParameter = True
                break

    # add parameter gridfile if not already existent
    if (not hasGridlineParameter):
        lines.append(newline)

    # write parameters to file
    with open("parameterMMDG.ini", "w") as file:
        file.write("\n".join(lines))


# reads error, run time and maximum edge length, i.e. the output from the
# excuted discontinuous Galerkin scheme
def readData():
    file = open("convergenceData", "r")

    # read data
    errorBulk = float(file.readline())
    errorInterface = float(file.readline())
    errorTotal = float(file.readline())
    hMax = float(file.readline())
    hMaxInterface = float(file.readline())
    timeTotal = float(file.readline())

    file.close()

    # return data
    return np.array([errorBulk, errorInterface, errorTotal, hMax, \
        hMaxInterface, timeTotal])


# determines the problem type from the ini file
def getProblemType ():
    # storage for problem type (1 to 6)
    problemType = None

    # read file content and get problem type
    with open("parameterMMDG.ini", "r") as file:
        lines = file.read().splitlines()

        for i in range(len(lines)):
            if (lines[i].startswith("problem")):
                problemType = int(lines[i].rstrip()[-1])
                break

    return problemType


# determines experimental order of convergence (EOC),
# data is assumed to be a numpy array with
# first column: error,
# second column: maximum edge length
def EOCloop (data):
    EOC = np.empty(data.shape[0] - 1)

    for i in range(len(EOC)):
        if data[i, 0] != 0:
            EOC[i] = log(data[i, 0] / data[i+1, 0]) / \
                log(data[i, 1] / data[i+1, 1])
        else:
            EOC[i] = -1.0*float('inf')
    # end for i

    return EOC


# === main ===

# create directory for the plots if necessary
if not exists("plots"):
    mkdir("plots")

# number of runs from which the run times are taken as mean value
numberOfMeans = 1

# names of the grid files sorted by refinement
gridfiles1246_2d = ["mmdg2_1", "mmdg2_5e-1", "mmdg2_2e-1", "mmdg2_8e-2", \
    "mmdg2_5e-2", "mmdg2_3e-2", "mmdg2_2e-2", "mmdg2_1e-2", "mmdg2_9e-3", \
    "mmdg2_8e-3", "mmdg2_6e-3", "mmdg2_4e-3"]
gridfiles3_2d = ["mmdg3_1", "mmdg3_5e-1", "mmdg3_2e-1", "mmdg3_8e-2", \
    "mmdg3_5e-2", "mmdg3_3e-2", "mmdg3_2e-2", "mmdg3_1e-2", "mmdg3_9e-3", \
    "mmdg3_8e-3"]
gridfiles5_2d = ["mmdg5_1", "mmdg5_5e-1", "mmdg5_2e-1", "mmdg5_8e-2", \
    "mmdg5_5e-2", "mmdg5_3e-2", "mmdg5_2e-2", "mmdg5_1e-2", "mmdg5_9e-3", \
    "mmdg5_8e-3"]
gridfiles1246_3d = ["mmdg2_1", "mmdg2_5e-1", "mmdg2_2e-1", "mmdg2_8e-2"]
gridfiles3_3d = ["mmdg3_1", "mmdg3_5e-1", "mmdg3_2e-1", "mmdg3_8e-2"]
gridfiles5_3d = ["mmdg5_1", "mmdg5_5e-1", "mmdg5_2e-1", "mmdg5_8e-2"]

gridfiles_2d = None
gridfiles_3d = None

# get problem type (integer from 1 to 6) from ini file
problemType = getProblemType()

# determine which gridfiles to use according to problem type
if problemType in [1, 2, 4, 6]:
    gridfiles_2d = gridfiles1246_2d
    gridfiles_3d = gridfiles1246_3d
elif problemType == 3:
    gridfiles_2d = gridfiles3_2d
    gridfiles_3d = gridfiles3_3d
elif problemType == 5:
    gridfiles_2d = gridfiles5_2d
    gridfiles_3d = gridfiles5_3d

#dim = 2
for dim in [2, 3]:

    gridfiles = gridfiles_2d if (dim == 2) else gridfiles_3d

    # storage for errors, run times and largestEdgeLength
    # column 0: error
    # column 1: total run time
    # column 2: largest edge length
    # different rows correspond to different numbers of grid elements
    data = np.zeros((len(gridfiles), 6))

    for i in range(len(gridfiles)):
        # write dgf file
        writeINI(gridfiles[i])

        # take mean value over numberOfMeans runs
        for j in range(numberOfMeans):
            # run executable
            call(join(".", "mmdg-" + str(dim) + "d"))

            # store data
            data[i, :] += readData()
        # end for j
    # end for i

    # divide by numberOfMeans
    data /= numberOfMeans

    # determine experimental order of convergence (EOC)
    EOCbulk = EOCloop(data[:, [0,3]])
    EOCinterface = EOCloop(data[:, [1,4]])
    EOCtotal = \
        EOCloop(np.column_stack([data[:,2], np.maximum(data[:,3], data[:,4])]))

    # print EOCs
    print("\n-------------------------------------------------------------\n")
    print("EOCs (" + str(dim) + "d):")
    print("\nbulk:")
    print(EOCbulk)
    print("\ninterface:")
    print(EOCinterface)
    print("\ntotal:")
    print(EOCtotal)
    print("\n=============================================================\n\n")

    # write data to file
    np.savetxt(join("plots", "mmdgAnalysis_" + str(dim) + "d"), data)
    np.savetxt(join("plots", "mmdgEOC_" + str(dim) + "d"), EOCtotal)
# end for dim

# remove output files
remove("convergenceData")
